//
// Copyright(C) Sergey A. Galin, 2008. http://sageshome.net/
//
// This program is distributed under terms of GNU General Public License,
// version 2 or higher, published by Free Software Foundation, at your option.
// See COPYING.txt for more information.
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "riff.h"

#define VERSION "0.1 alpha 1"

const int FRAMES_PER_SECOND=75;
const int MAX_FRAME=74;
const int FRAME_SIZE=44100/FRAMES_PER_SECOND; // ==588 - frame length in samples


struct TCUEEntry
{
	static const int MAX_CUE_ENTRIES = 10000;
	char *title, *performer, *trackNo, *type;
	// Negative values in end position indicate that
	// the entry continues to the end of the stream.
	int startMin, startSec, startFrame, endMin, endSec, endFrame;

	void Clear()
	{
		title = "";
		performer = "";
		trackNo = "";
		type = "";
		startMin = 0;
		startSec = 0;
		startFrame = 0;
		endMin = -1;
		endSec = -1;
		endFrame = -1;
	}

	void ParseTimeString(const char *str){
		if( sscanf(str, "%d:%d:%d", &startMin, &startSec, &startFrame)!=3 )
			printf("Warning: parsing time string failed: ''.\n", str);
	}

	long StartInSamples() const{
		return (((long)startMin*60+startSec)*FRAMES_PER_SECOND+
				startFrame)*FRAME_SIZE;
	}

	// May return -1 if end was not set
	long EndInSamples() const{
		if( endMin>=0 && endSec>=0 && startFrame>=0 )
			return (((long)endMin*60+endSec)*FRAMES_PER_SECOND+
				endFrame+1)*FRAME_SIZE-1;
		else
			return -1;
	}
	
	void Echo(){
		printf("***** Track #%s of type '%s'\n"
		       "*****     Title: '%s'\n"
		       "*****     Performer: '%s'\n"
		       "*****     Location: %d:%d:%d..%d:%d:%d (%ld..%ld)\n\n",
		       trackNo, type, title, performer,
		       startMin, startSec, startFrame, endMin, endSec, endFrame,
		       StartInSamples(), EndInSamples());
	}

	
};


// Looks in current line only!
// Returns -1 if keyword not found, otherwise.pointer which
// is set AFTER the keyword.
char *FindKeywordInLine(char *haystack, const char *needle)
{
	while( isspace(*haystack)){
		if( !(*haystack) || (*haystack=='\n') )
			return NULL;
		haystack++;
	}
	if( !(*haystack) || (*haystack=='\n') )
		return NULL;
	int len = strlen(needle);
	if( !strncmp( haystack, needle, len ) )
		return haystack + len;
	return NULL;
}

// This function alters 'str', adding 0 character after the value,
// so add strlen(value)+1 to str to extract more values.
// Note that if no value is found, an empty string is returned.
char *ExtractValue(char *str)
{
	while( isspace(*str) ){
		if( !(*str) || (*str)=='\n' )
			return "";
		str++;
	}
	// Quoted value
	if( (*str)=='\"' ){
		char *ret = ++str;
		while( (*str)!='\"' && (*str)!='\n' && (*str) )
			str++;
		(*str) = 0;
		return ret;
	}
	// Unquoted value, also ends when space character found.
	char *ret = str;
	while( (*str)!='\n' && !isspace(*str) && (*str) )
		str++;
	*str = 0;
	return ret;
}

// This functions parses CUE file fully loaded into 'cue', and
// writes found values into the variables passed.
// No memory for lines is allocated; the variables are pointed
// directly to their values in 'cue'. Zero bytes are written
// into 'cue' to indicate ends of values.
// 'entries' must be at least MAX_CUE_ENTRIES elements.
// Returns number of entries found.
int ParseCUE(char *cue, char **wavfile,
	char **performer, char **album, TCUEEntry *entries)
{
	int nEntries = 0;
	*wavfile = "";
	*performer = "";
	*album = "";
	memset(entries, 0, sizeof(TCUEEntry)*TCUEEntry::MAX_CUE_ENTRIES);
	printf("Parsing CUE sheet...\n");

	// Parsing header
	char *line = cue;
	bool bTrackFound = false;
	while( *line ){
		char *data = NULL;
		// ExtractValue() destroys lines, so without this
		// we cannot tell where the next line is:
		char *nextline = strchr( line, '\n' );
		if( (data=FindKeywordInLine(line, "PERFORMER"))!=NULL ){
			*performer = ExtractValue( data );
			// line += strlen(*performer)+1;
			printf("*** Performer: %s\n", *performer);
		}else if( (data=FindKeywordInLine(line, "TITLE"))!=NULL ){
			*album = ExtractValue( data );
			// line += strlen(*album)+1;
			printf("*** Album Title: %s\n", *album);
		}else if( (data=FindKeywordInLine(line, "FILE"))!=NULL ){
			*wavfile = ExtractValue( data );
			// line += strlen(*wavfile)+1;
			printf("*** Data File: %s\n", *wavfile);
		}else if( (data=FindKeywordInLine(line, "TRACK"))!=NULL ){
			bTrackFound = true;
			
			// First entry parsing
			// NOTE THAT THE CODE IS DUPLICATED BELOW!
			entries[nEntries].Clear();
			entries[nEntries].performer = *performer;
			entries[nEntries].trackNo = ExtractValue( data );
			data+=strlen( entries[nEntries].trackNo )+1;
			if( !(*(data+1)) ) // Error!
				return 0;
			data++;
			entries[nEntries].type = ExtractValue( data );
			nEntries++;
			
			if( nextline==NULL )
				return 0;
			line = nextline+1; // Step over the '\n'

			break; // Continue to parsing entries
		}
		// Seeking to next line.
		if( nextline==NULL )
			return 0;
		line = nextline+1; // Step over the '\n'
	}

	if( !bTrackFound )
		return 0;
	
	// Parsing entries.
	// Upon entry to each cycle, 'line' points to line where
	// TRACK entry is located.
	while( *line ){
		char *data = NULL;
		char *nextline = strchr( line, '\n' );

		if( (data=FindKeywordInLine(line, "TRACK"))!=NULL ){
			if( nEntries==TCUEEntry::MAX_CUE_ENTRIES ){
				fprintf(stderr, "Max number of entries reached!\n");
				return nEntries;
			}
		
			// NOTE THAT THE CODE IS DUPLICATED ABOVE!
			entries[nEntries].Clear();
			entries[nEntries].performer = *performer;
			entries[nEntries].trackNo = ExtractValue( data );
			data+=strlen( entries[nEntries].trackNo )+1;
			if( !(*(data+1)) ) // Error!
				return 0;
			data++;
			entries[nEntries].type = ExtractValue( data );
			nEntries++;			
		}else if( (data=FindKeywordInLine(line, "TITLE"))!=NULL ){
			entries[nEntries-1].title = ExtractValue( data );
			// line += strlen(*entries[nEntries-1].title)+1;
		}else if( (data=FindKeywordInLine(line, "PERFORMER"))!=NULL ){
			entries[nEntries-1].performer = ExtractValue( data );
			// line += strlen(*entries[nEntries-1].perfomer)+1;
		}else if( (data=FindKeywordInLine(line, "INDEX"))!=NULL ){
			char *number=ExtractValue( data );
			int inumber;
			if( sscanf(number, "%d", &inumber)==1 ){ // Parsed successfully
				// Alwayws using only INDEX 01, which determines beginning of the
				// track after pregap.
				if( inumber==1 ){
					data+=strlen(number)+2;
					char *time=ExtractValue( data );
					// printf("Number: '%s'; Extracted time string: '%s'\n", number, time);
					entries[nEntries-1].ParseTimeString( time );
					// Dont't forget to update end time of the previous entry
					if( nEntries>1 ){
						int min=entries[nEntries-1].startMin,
							sec=entries[nEntries-1].startSec,
							frame=entries[nEntries-1].startFrame;
						// Going one frame back
						if( frame>0 )
							frame--;
						else{
							if( sec>0 ){
								sec--;
								frame=MAX_FRAME;
							}else{
								if( min==0 ){
									printf("Warning: zero start time for non-first track!\n");
									continue;
								}
								min--;
								sec=59;
								frame=MAX_FRAME;
							}
						}
						entries[nEntries-2].endMin = min;
						entries[nEntries-2].endSec = sec;
						entries[nEntries-2].endFrame = frame;
					}
				}
			}else
				printf("Warning: failed to parse INDEX number.\n");
		}

		

		// Seeking to next line.
		if( nextline==NULL )
			return 0;
		line = nextline+1; // Step over the '\n'		
	}

	for(int i=0; i<nEntries; i++)
		entries[i].Echo();


	return nEntries;
}

const char* Fourcc2Str( FOURCC code )
{
	static char buf[5]={0,0,0,0,0};
	*((FOURCC*)buf)=code;
	return buf;
}

void ExplodeFile(const char *wavfile,
	char *performer, char *album, TCUEEntry *entries, size_t nEntries)
{
	TRIFFWavHeader header;
	memset( &header, 0, sizeof(header) );
	FILE *fp=fopen(wavfile, "rb");
	if( !fp ){
		printf("File cannot be opened: '%s'\n", wavfile);
		return;
	}
	if( fread( &header, sizeof(header), 1, fp )!=1 ){
		printf("Error reading PCM WAV header.\n");
		return;
	}

	printf("** RIFF Signature: '%s'\n", Fourcc2Str(header.riffheader));
	printf("** WAVE Signature: '%s'\n", Fourcc2Str(header.WAVEsig));
	printf("** Data Signature: '%s'\n", Fourcc2Str(header.dataheader));
	printf("** File length-8: %u\n", header.WAVElen);
	printf("** FMT length-8: %u\n", header.fmtlen);
	printf("** Format Tag: %d\n", int(header.FormatTag));
	printf("** Channels: %d\n", int(header.Channels));
	printf("** Samples Per Second: %u\n", header.SamplesPerSec);
	printf("** Bytes Per Second: %u\n", header.AvgBytesPerSec);
	printf("** Block Alignment: %d\n", int(header.BlockAlign));
	printf("** Bits Per Sample: %d\n", int(header.BitsPerSample));
	printf("** Data Length: %u\n", header.datalen);
	

	if( !header.IsValidPCMWav() ){
		printf("The file is not a simple PCM WAV.\n"
		  "Sorry, I'm too stupid to handle that.\n");
		return;
	}
	if( header.Channels!=2 || header.SamplesPerSec!=44100 ){
		printf("The file is not stereo/44100Hz.\n"
		  "Sorry, I'm too stupid to handle that.\n");
		return;
	}
	// header will be reused later for writing files
	long size = header.datalen;
	const int sampleSize = 4; // 4 bytes for 2 channels x 16 bit
	long totalSamples = size/sampleSize;
	printf("Data size: %ld bytes / %ld samples.\n", size, totalSamples);

	char fileName[1024]; // !!!!!!!!!!!!!!!!!!!!!!!!!!! FIX THAT
	for(size_t i=0; i<nEntries; i++){
		fileName[0]=0;
		int trackNo=0;
		sscanf(entries[i].trackNo, "%d", &trackNo);
		// char *title, *performer, *trackNo, *type;
		if( trackNo )
			sprintf(fileName, "%02d ", trackNo);
		strcat(fileName, entries[i].title);
		strcat(fileName, ".wav");
		for(char *ptr=fileName; *ptr; ptr++){
			if( !isalpha(*ptr) && !isdigit(*ptr) &&
				*ptr!=' ' && *ptr!='-' && *ptr!='_' && *ptr!='.' )
				*ptr = '_';
		}
		printf("Extracting entry #%d into '%s'...\n", int(i), fileName);
		long start=entries[i].StartInSamples(), end=entries[i].EndInSamples();
		if( end<start )
			end = totalSamples;
		header.datalen = DWORD((end-start)*4);
		long toCopy = long(header.datalen);
		header.WAVElen = DWORD(sizeof(header)-8+header.datalen);
		FILE *fout=fopen( fileName, "wb" );
		if( !fout ){
			printf("Error: failed to open file for writing.\n");
			exit(255);
		}
		if( fwrite( &header, sizeof(header), 1, fout )!=1 ){
			printf("Error: failed to write PCM WAV header!\n");
			exit(255);
		}

		const int bufferSize=1024*1024; // 1 meg should be enough for everyone (C)
		char buffer[ bufferSize ];
		fseek( fp, start*4+sizeof(header), SEEK_SET );
		long copied=0;
		while( toCopy>0 ){
			long toCopyNow = toCopy;
			if( toCopyNow > bufferSize )
				toCopyNow = bufferSize;
			if( fread( buffer, toCopyNow, 1, fp )!=1 ){
				printf("Error: short read!\n");
				exit(255);
			}
			if( fwrite( buffer, toCopyNow, 1, fout )!=1 ){
				printf("Error: short write!\n");
				exit(255);
			}
			copied += toCopyNow;
			toCopy -= toCopyNow;
			printf("%ld bytes copied...  \r", copied);
			fflush( stdout );
		}
		printf("\n");
		
		fclose(fout);
	}
	
			
	fclose(fp);
}



int main(int argc, char *argv[])
{
	printf("CUE blow up - a PCM WAV+CUE extractor, v." VERSION 
	         " Compiled: " __DATE__ " " __TIME__ "\n"
	       "Written by Sergey A. Galin, 2008. Homepage: http://sageshome.net/\n"
	       "This program is distributed under GNU General Public License,\n"
           "version 2.0 or higher, published by Free Software Foundation.\n"
	       );

	if( sizeof(DWORD)!=4 || sizeof(WORD)!=2 ){
		printf("The program was not configured properly!!\n");
		return -1;
	}

	if( argc!=2 ){
		printf("USAGE: cueblowup <cue file>\n");
		return 1;
	}

	char *fn=argv[2];

	FILE *fp = fopen(fn, "rt");
	if( !fp ){
		printf("Couldn't open CUE file.\n");
		return 1;
	}

	fseek( fp, 0, SEEK_END );
	long sz = ftell( fp );
	fseek( fp, 0, SEEK_SET );
	printf("File size: %ld bytes.\n", sz);
	if( sz<=0 ){
		printf("CUE sheet is empty.\n");
		return 1;
	}
	char *cuedata = new char[ sz+1 ];
	if( !cuedata ){
		printf("Failed to allocate memory for CUE sheet.\n");
		return 1;
	}
	memset( cuedata, 0, sz+1 );
	if( fread( cuedata, 1, sz, fp )!=sz ){
		printf("File read failure.\n");
		return 1;	
	}
	fclose(fp);

	char *wavfile, *performer, *album;
	TCUEEntry entries[TCUEEntry::MAX_CUE_ENTRIES];
	int nEntries;

	nEntries = ParseCUE( cuedata, &wavfile, &performer, &album, entries );
	if( nEntries<=0 ){
		printf("No entries found.\n");
		delete cuedata;
		return 1;	
	}

	ExplodeFile( wavfile, performer, album, entries, nEntries );
	
	delete cuedata;

}

